#ifndef Agre_BASICS_H
#define Agre_BASICS_H

//#include <mitsuba/core/platform.h>
//#include <mitsuba/core/fwd.h>

#include <mitsuba/render/sampler.h>
#include"furI.h"
#include"net.h"
#include <iostream>
#include <cstdio>
#include <assert.h>
#include "helpers.h"
#include "spline.h"
using namespace std;
using namespace mitsuba;
MTS_NAMESPACE_BEGIN
class AggregatedSolver {
public:
	//parameters for evaluating the inherited part
	Float eta;
	Vector3 sigma_a;
	Float alpha;
	Float beta_m, beta_n;
	Float kappa;
	Float sigma_a_m, sigma_s_m;
	Float g;
	Float layers;
	Float Ti;
	Sampler *sampler;
	//parameter A_d for evaluating the direct part
	/*Float rho;
	int fibernum;*/
	Float A_d;
	//output parameters for the param network for evaluating the extended part.
	Float Iscale[3];
	Float theta_top[3];
	Float theta_middle[3];
	Float theta_bottom[3];
	Float beta_top[3];
	Float beta_middle[3];
	Float beta_bottom[3];
	Vector3 A_top;
	Vector3 A_middle;
	Vector3 A_bottom;
	Spline EN[3];

public:

	AggregatedSolver(Float eta, Vector3 sigma_a, Float alpha, Float beta_m, Float beta_n,
		Float kappa, Float sigma_a_m, Float sigma_s_m, Float g, Float layers, Float rho, int fibernum, Float theta_i, Sampler *sampler_ = NULL) :
		eta(eta), sigma_a(sigma_a), alpha(alpha), beta_m(beta_m), beta_n(beta_n),
		kappa(kappa), sigma_a_m(sigma_a_m), sigma_s_m(sigma_s_m), g(g), layers(layers) {
		if (sampler_)
			sampler = sampler_;
		A_d = trans(rho, fibernum);
		FURNetoutput out;
		//infer the lightweight neural network when initialize the aggregated fiber
		out = forward(eta, sigma_a, alpha, beta_m, beta_n, kappa, sigma_a_m, sigma_s_m, g, layers, rho, fibernum, theta_i);
		//infer the RGB seperately due to sigma_a
		for (int i = 0; i < 3; i++)
		{
			Iscale[i] = out[0 * 3 + i];
			theta_top[i] = out[1 * 3 + i];
			theta_middle[i] = out[2 * 3 + i];
			theta_bottom[i] = out[3 * 3 + i];
			beta_top[i] = out[4 * 3 + i];
			beta_middle[i] = out[5 * 3 + i];
			beta_bottom[i] = out[6 * 3 + i];
			A_top[i] = out[7 * 3 + i];
			A_middle[i] = out[8 * 3 + i];
			A_bottom[i] = out[9 * 3 + i];
			Float y0 = out[10 * 3 + i];
			Float y1 = out[11 * 3 + i];
			Float y2 = out[12 * 3 + i];
			Float y3 = out[13 * 3 + i];
			Float y4 = out[14 * 3 + i];
			EN[i].set_points(y0, y1, y2, y3, y4);
		}		
	}

	Float getScatteredN(Float sigma_s, Float h, Float g, Float phi) const {
		if (h < 0.0) {
			h = -h;
			phi = -phi;
		}

		Float binSigma = sigma_s / MAX_SCATTERING_INNER * NUM_SCATTERING_INNER;
		Float binH = h / MAX_H * NUM_H;
		Float binG = g / MAX_G * NUM_G;
		Float binPhi;
		if (phi < 0)
			binPhi = (-phi) / (2.0 * M_PI) * NUM_BINS;
		else
			binPhi = (2.0 * M_PI - phi) / (2.0 * M_PI) * NUM_BINS;
		return interpScatteredN(binSigma, binH, binG, binPhi);
	}

	Float randomnum()
	{
		Float random = sampler->next1D();
		return random;
	}

	Float getScatteredDist(Float sigma_s, Float h, Float g, Float phi) const {
		if (h < 0.0) {
			h = -h;
			phi = -phi;
		}
		Float binSigma = sigma_s / MAX_SCATTERING_INNER * NUM_SCATTERING_INNER;
		Float binH = h / MAX_H * NUM_H;
		Float binG = g / MAX_G * NUM_G;
		// if (phi < 0) phi += 2.0 * M_PI;
		Float binPhi;
		if (phi < 0)
			binPhi = (-phi) / (2.0 * M_PI) * NUM_BINS;
		else
			binPhi = (2.0 * M_PI - phi) / (2.0 * M_PI) * NUM_BINS;
		return interpScatteredDist(binSigma, binH, binG, binPhi);
	}

	Float theta_cone_R(Float theta_i, Float alpha, Float phi) const {
		Float u = cos(phi / 2.0) * cos(alpha) * cos(theta_i) + sin(alpha) * sin(theta_i);
		Float t = sin(theta_i) - 2.0 * sin(alpha) * u;
		return -asinSafe(t);
	}

	Float beta_m_cone_R(Float beta_m, Float phi) const {
		return std::max(beta_m * std::sqrt(2.0) * cos(phi / 2.0), degToRad(0.02));
		//        return beta_m;
	}

	Float Mu(int p, Float theta_i, Float theta_r, Float phi) const {
		Float alpha_p, beta_m_p;

		switch (p) {
		case 0: // R
			alpha_p = alpha;
			beta_m_p = beta_m;
			return G(theta_r - theta_cone_R(theta_i, alpha_p, phi), beta_m_cone_R(beta_m_p, phi));
			break;
		case 1: // TT
			alpha_p = -alpha / 2.0;
			beta_m_p = beta_m; 
			break;
		case 2: // TRT
			alpha_p = -alpha / 2.0 * 3.0;
			beta_m_p = beta_m / 2.0 * 3.0;
			break;
		}

		return G(theta_r - (-theta_i + 2.0 * alpha_p), beta_m_p);
	}

	Float Ms(Float theta_i, Float theta_r, Float phi) const {
		if (theta_i < 0.0) {
			theta_i = -theta_i;
			theta_r = -theta_r;
		}

		Float binSigma = sigma_s_m * kappa / MAX_SCATTERING_INNER * NUM_SCATTERING_INNER;
		Float binThetaI = theta_i / MAX_THETA * NUM_THETA;
		Float binG = g / MAX_G * NUM_G;
		Float binThetaRBack = (theta_r + M_PI / 2.0) / M_PI * (NUM_BINS / 2 - 1);
		Float binThetaRFront = NUM_BINS - 1 - binThetaRBack;
		Float wBack = (M_PI - std::fabs(phi)) / M_PI;
		Float wFront = 1.0 - wBack;
		return interpScatteredM(binSigma, binThetaI, binG, binThetaRBack) * wBack +
			interpScatteredM(binSigma, binThetaI, binG, binThetaRFront) * wFront;

	}

	Float Me(int p, Float theta_i, Float theta_r, Float phi) const{
		Float alpha_p, beta_m_p;

		switch (p) {
		case 0: // top
			alpha_p = Vector3(theta_top).length();
			beta_m_p = Vector3(beta_top).length();
		case 1: // middle
			alpha_p = Vector3(theta_middle).length();
			beta_m_p = Vector3(beta_middle).length();
			break;
		case 2: // bottom
			alpha_p = Vector3(theta_bottom).length();
			beta_m_p = Vector3(beta_bottom).length();
			break;
		}
		return G(theta_r - (alpha_p), beta_m_p);
	
	}

	Float Md( Float theta_i, Float theta_r) const{		
		return G(theta_i + theta_r, Rd);

	}

	Vector3 Au(int p, Float h, Float theta_i, Float eta_prime) const {
		Float gamma_i = asinSafe(h);
		Float gamma_t = asinSafe(h / eta_prime);
		Float f = fresnelCuticle(gamma_i, 1.0, eta_prime, layers);
		if (p == 0) {
			return Vector3(f, f, f);
		}
		else {
			Float theta_t = asinSafe(1.0 / eta * sin(theta_i));
			Vector3 sigma_a_prime = sigma_a / cos(theta_t);
			Float sigma_a_m_prime = sigma_a_m / cos(theta_t);
			Float sigma_s_m_prime = sigma_s_m / cos(theta_t);
			Float sigma_e_m_prime = sigma_a_m_prime + sigma_s_m_prime;

			// Travelling distance in the cortex and medulla, respectively.
			Float d1 = cos(gamma_t), d2 = 0.0;
			Float h_prime = h / eta_prime;
			if (std::fabs(h_prime) < std::fabs(kappa))
				d2 = std::sqrt(kappa * kappa - h_prime * h_prime);
			d1 -= d2;
			return Float(std::pow(1.0 - f, 2.0)) *
				Float(std::pow(f, p - 1.0)) *
				pow3(exp3(-2.0 * d1 * sigma_a_prime) *
				std::exp(-2.0 * d2 * sigma_e_m_prime), p);
		}
	}

	Float Du(Float phi, Float sigma) const {
		Float retVal = 0.0;
		for (int i = -1; i <= 1; i++)
			retVal += G(phi + i * 2.0 * M_PI, sigma);
		return retVal;
	}

//with h near-field calculation
	Vector3 Nu(int p, Float h, Float phi, Float theta_i, Float eta_prime) const {
		Float gamma_i = asinSafe(h);
		Float gamma_t = asinSafe(h / eta_prime);
		Vector3 attenuation = Au(p, h, theta_i, eta_prime);
		attenuation[0] *= Iscale[0];
		attenuation[1] *= Iscale[1];
		attenuation[2] *= Iscale[2];
		Float Phi = 2.0 * p * gamma_t - 2.0 * gamma_i + p * M_PI;
		Float beta_n_p_sqr = beta_n * beta_n * (p + 1);
		Float beta_n_p = std::sqrt(beta_n_p_sqr);
		return attenuation * Du(Phi - phi, beta_n_p);
	}

	Vector3 Nu(int p, Float h, Float phi, Float theta_i) const {
		Float eta_prime = std::sqrt(eta * eta - sin(theta_i) * sin(theta_i)) / cos(theta_i);
		return Nu(p, h, phi, theta_i, eta_prime);
	}

	Vector3 As(int p, Float h, Float phi, Float theta_i, Float eta_prime) const {
		assert(p > 0);
		Float gamma_i = asinSafe(h);
		Float gamma_t = asinSafe(h / eta_prime);
		Float f = fresnelCuticle(gamma_i, 1.0, eta_prime, layers);
		Float theta_t = asinSafe(1.0 / eta * sin(theta_i));
		Vector3 sigma_a_prime = sigma_a / cos(theta_t);
		Float sigma_a_m_prime = sigma_a_m / cos(theta_t);
		Float sigma_s_m_prime = sigma_s_m / cos(theta_t);
		Float sigma_e_m_prime = sigma_a_m_prime + sigma_s_m_prime;
		Float d1 = cos(gamma_t), d2 = 0.0;
		Float h_prime = h / eta_prime;
		if (std::fabs(h_prime) >= std::fabs(kappa))
			return Vector3(0.0, 0.0, 0.0);
		d2 = std::sqrt(kappa * kappa - h_prime * h_prime);
		d1 -= d2;
		Vector3 retVal = (1.0 - f) * std::pow(f, p - 1.0) *
			mul3(exp3(-(2 * p - 1) * d1 * sigma_a_prime),
			exp3(-(1.0 - kappa) * sigma_a_prime)) *
			std::exp(-2 * d2 * (p - 1) * sigma_e_m_prime);
		Float phi_query = phi - (gamma_t - gamma_i) - (M_PI + 2.0 * gamma_t) * (p - 1);
		regularizePhi(phi_query);
		Float avgDist = kappa * getScatteredDist(sigma_s_m_prime * kappa, h_prime / kappa, g, phi_query);
		retVal = retVal * std::exp(-avgDist * sigma_a_m_prime)*(Vector3(1.) -
			pow3(exp3(-2.0 * d1 * sigma_a_prime) *
			std::exp(-2.0 * d2 * sigma_e_m_prime), p));
		return retVal;
	}

	Float Ds(int p, Float h, Float phi, Float theta_i, Float eta_prime) const {
		assert(p > 0);
		Float gamma_i = asinSafe(h);
		Float gamma_t = asinSafe(h / eta_prime);
		Float theta_t = asinSafe(1.0 / eta * sin(theta_i));
		Float sigma_s_m_prime = sigma_s_m / cos(theta_t);
		Float h_prime = h / eta_prime;
		Float phi_query = phi - (gamma_t - gamma_i) - (M_PI + 2.0 * gamma_t) * (p - 1);
		regularizePhi(phi_query);
		return getScatteredN(sigma_s_m_prime * kappa, h_prime / kappa, g, phi_query);
	}

	Vector3 Ns(int p, Float h, Float phi, Float theta_i, Float eta_prime) const {
		if (p == 0)
			return Vector3(0.0, 0.0, 0.0);
		Float h_prime = h / eta_prime;
		if (std::fabs(h_prime) >= std::fabs(kappa))
			return Vector3(0.0, 0.0, 0.0);
		Vector3 attenuation = As(p, h, phi, theta_i, eta_prime);
		attenuation[0] *= Iscale[0];
		attenuation[1] *= Iscale[1];
		attenuation[2] *= Iscale[2];
		return attenuation * Ds(p, h, phi, theta_i, eta_prime);
	}

	Vector3 Ns(int p, Float h, Float phi, Float theta_i) const {
		Float eta_prime = std::sqrt(eta * eta - sin(theta_i) * sin(theta_i)) / cos(theta_i);
		return Ns(p, h, phi, theta_i, eta_prime);
	}
	// Choose a lobe to sample.

	void chooseLobe(Float theta_i, Float h, int &lobe, Float &pdfLobe) const {
		Float eta_prime = std::sqrt(eta * eta - sin(theta_i) * sin(theta_i)) / cos(theta_i);
		Float ASpec[9];
		ASpec[0] = Au(0, h, theta_i, eta_prime).length();
		ASpec[1] = Au(1, h, theta_i, eta_prime).length();
		ASpec[2] = Au(2, h, theta_i, eta_prime).length();
		ASpec[3] = As(1, h, M_PI / 2.0, theta_i, eta_prime).length();
		ASpec[4] = As(2, h, M_PI / 2.0, theta_i, eta_prime).length();
		ASpec[5] = A_top.length();
		ASpec[6] = A_middle.length();
		ASpec[7] = A_bottom.length();
		ASpec[8] = A_d;
		Float sumPdf = 0;
		for (int i = 0; i < 9; i++)
			sumPdf += ASpec[i];
		if (sumPdf == 0.0) {
			lobe = -1;
			pdfLobe = 0.0;
			return;
		}
		for (int i = 0; i < 9; i++)
			ASpec[i] /= sumPdf;
		int choice = 8;
		Float randNum = sampler->next1D();
		for (int i = 0; i < 9; i++) {
			randNum -= ASpec[i];
			if (randNum <= 0.0) {
				choice = i;
				break;
			}
		}
		lobe = choice;
		pdfLobe = ASpec[choice];
	}

	void sampleMr(Float theta_i, Float phi, Float &theta_o, Float &pdfM, Float &weightM) const {
		Float theta_cone = theta_cone_R(theta_i, alpha, phi);
		Float beta_cone = beta_m_cone_R(beta_m, phi);
		theta_o = sampleGaussian(theta_cone, beta_cone, sampler->next1D(), sampler->next1D());
		pdfM = G(theta_o - theta_cone, beta_cone);
		weightM = 1.0;
	}

	void sampleMu(int p, Float theta_i, Float &theta_o, Float &pdfM, Float &weightM) const {
		Float alpha_p, beta_m_p;
		switch (p) {
		case 1: // TT
			alpha_p = -alpha / 2.0;
			beta_m_p = beta_m; // / 2.0;
			break;
		case 2: // TRT
			alpha_p = -alpha / 2.0 * 3.0;
			beta_m_p = beta_m / 2.0 * 3.0;
			break;
		}
		theta_o = sampleGaussian(-theta_i + 2.0 * alpha_p, beta_m_p, sampler->next1D(), sampler->next1D());
		pdfM = G(theta_o - (-theta_i + 2.0 * alpha_p), beta_m_p);
		weightM = 1.0;
	}

	void sampleMe(int p, Float theta_i, Float &theta_o, Float &pdfM, Float &weightM) const
	{
		Float alpha_p, beta_m_p;
		switch (p-5) {
		case 0: // top
			alpha_p = Vector3(theta_top).length();//different absorptions have the same distribution shape 
			beta_m_p = Vector3(beta_top).length();
		case 1: // middle
			alpha_p = Vector3(theta_middle).length();
			beta_m_p = Vector3(beta_middle).length(); // / 2.0;
			break;
		case 2: // bottom
			alpha_p = Vector3(theta_bottom).length();
			beta_m_p = Vector3(beta_bottom).length();
			break;
		}
		theta_o = sampleGaussian(alpha_p, beta_m_p, sampler->next1D(), sampler->next1D());
		pdfM = G(theta_o - (alpha_p), beta_m_p);
		weightM = 1.0;
	}

	void sampleMd( Float theta_i, Float &theta_o, Float &pdfM, Float &weightM) const
	{	
		theta_o = sampleGaussian(-theta_i, Rd, sampler->next1D(), sampler->next1D());
		pdfM = G(theta_o + (theta_i), Rd);
		weightM = 1.0;
	}

	void sampleNu(int p, Float h, Float theta_i, Float phi_i, Float &phi_o, Float &pdfN, Vector3 &weightN) const {
		Float eta_prime = std::sqrt(eta * eta - sin(theta_i) * sin(theta_i)) / cos(theta_i);
		Float gamma_i = asinSafe(h);
		Float gamma_t = asinSafe(h / eta_prime);
		Vector3 attenuation = Au(p, h, theta_i, eta_prime);
		Float Phi = 2.0 * p * gamma_t - 2.0 * gamma_i + p * M_PI;
		Float beta_n_p_sqr = beta_n * beta_n * (p + 1);
		Float beta_n_p = std::sqrt(beta_n_p_sqr);
		Float phi = sampleGaussian(Phi, beta_n_p, sampler->next1D(), sampler->next1D());
		phi_o = phi_i + phi;
		pdfN = G(Phi - phi, beta_n_p);
		weightN = attenuation;
	}

	void sampleNe(int p, Float h, Float theta_i, Float phi_i, Float &phi_o, Float &pdfN, Vector3 &weightN) const {
		Float phi = (sampler->next1D() * 2.0 - 1.0) * M_PI;
		pdfN = 1.0 / (2.0 * M_PI);
		phi_o = phi_i + phi;
		weightN = max(discardLarger(Ne(p-5, h, phi), 5.0), 0.0) / pdfN;
	}

	void sampleNd(int p, Float h, Float theta_i, Float phi_i, Float &phi_o, Float &pdfN, Vector3 &weightN) const {
		Vector3 attenuation = Vector3(A_d);
		Float phi = sampleGaussian(M_PI, Rd, sampler->next1D(), sampler->next1D());
		phi_o = phi_i + phi;
		pdfN = G(phi, Rd);
		weightN = attenuation;
	}

	void sampleMs(Float theta_i, Float phi, Float &theta_o, Float &pdfM, Float &weightM) const {
		theta_o = asinSafe(sampler->next1D() * 2.0 - 1.0);
		pdfM = 0.5 * cos(theta_o);
		weightM = Ms(theta_i - alpha, theta_o - alpha, phi) / pdfM;
	}

	void sampleNs(int p, Float h, Float theta_i, Float phi_i, Float &phi_o, Float &pdfN, Vector3 &weightN) const {
		Float phi = (sampler->next1D() * 2.0 - 1.0) * M_PI;
		pdfN = 1.0 / (2.0 * M_PI);
		phi_o = phi_i + phi;
		weightN = max(discardLarger(Ns(p, h, phi, theta_i), 5.0), 0.0) / pdfN;
	}

	// Sample the chosen lobe. Return theta_o, phi_o, and corresponding sampling weights.
	void sampleLobe(int lobe, Float h, Float theta_i, Float phi_i, Float &theta_o, Float &pdfM, Float &weightM, Float &phi_o, Float &pdfN, Vector3 &weightN) const {
		if (lobe == 0) {
			// R lobe.
			sampleNu(lobe, h, theta_i, phi_i, phi_o, pdfN, weightN);
			Float phi = phi_o - phi_i;
			regularizePhi(phi);
			sampleMr(theta_i, phi, theta_o, pdfM, weightM);
		}
		else if (lobe <= 2) {
			// Unscattered TT and TRT lobes.
			sampleNu(lobe, h, theta_i, phi_i, phi_o, pdfN, weightN);
			sampleMu(lobe, theta_i, theta_o, pdfM, weightM);
		}
		else if(lobe<=4){
			// Scattered TT and TRT lobes.
			sampleNs(lobe - 2, h, theta_i, phi_i, phi_o, pdfN, weightN);
			Float phi = phi_o - phi_i;
			regularizePhi(phi);
			sampleMs(theta_i, phi, theta_o, pdfM, weightM);
		}
		else if (lobe <= 7)
		{
			sampleNe(lobe, h, theta_i, phi_i, phi_o, pdfN, weightN);
			sampleMe(lobe, theta_i, theta_o, pdfM, weightM);
		}
		else
		{
			sampleNd(lobe, h, theta_i, phi_i, phi_o, pdfN, weightN);
			sampleMd(theta_i, theta_o, pdfM, weightM);
		}
	}	

	// Longitudinal PDF of a chosen lobe
	inline Float pdfMLobe(int lobe, Float theta_i, Float theta_o, Float phi) const {
		if (lobe <= 2) {
			// Unscattered R, TT and TRT lobes.
			regularizePhi(phi);
			return Mu(lobe, theta_i, theta_o, phi);
		}
		else if (lobe<=4){
			// Scattered TT and TRT lobes.
			//            return 1.0 / M_PI;
			return 0.5 * cos(theta_o);
		}
		else if (lobe<=7){
			// Extended part  t m b
			return Me(lobe-5, theta_i, theta_o, phi);
		}
		else
		{
			// Direct transport
			return Md(theta_i, theta_o);
		}
	}

	// Azimuthal PDF of a chosen lobe
	inline Float pdfNLobe(int lobe, Float h, Float theta_i, Float phi) const {
		if (lobe <= 2) {
			// Unscattered R, TT and TRT lobes.
			int p = lobe;
			Float eta_prime = std::sqrt(eta * eta - sin(theta_i) * sin(theta_i)) / cos(theta_i);
			Float gamma_i = asinSafe(h);
			Float gamma_t = asinSafe(h / eta_prime);
			Float Phi = 2.0 * p * gamma_t - 2.0 * gamma_i + p * M_PI;
			Float beta_n_p_sqr = beta_n * beta_n * (p + 1);
			Float beta_n_p = std::sqrt(beta_n_p_sqr);
			Float phiDiff = Phi - phi;
			regularizePhi(phiDiff);
			return Du(phiDiff, beta_n_p);
		}
		else if (lobe == 7)
		{
			return G(normailzephi(phi-M_PI), Rd);
		}
		else {
			// Scattered TT and TRT lobes & extended top middle and bottom lobes.
			return 1.0 / (2.0 * M_PI);
		}
	}

	Float pdfMN(Float h, Float theta_i, Float theta_o, Float phi) const {
		Float eta_prime = std::sqrt(eta * eta - sin(theta_i) * sin(theta_i)) / cos(theta_i);
		Float ASpec[9];
		ASpec[0] = Au(0, h, theta_i, eta_prime).length();
		ASpec[1] = Au(1, h, theta_i, eta_prime).length();
		ASpec[2] = Au(2, h, theta_i, eta_prime).length();
		ASpec[3] = As(1, h, M_PI / 2.0, theta_i, eta_prime).length();
		ASpec[4] = As(2, h, M_PI / 2.0, theta_i, eta_prime).length();
		//extended part
		ASpec[5] = A_top.length();
		ASpec[6] = A_middle.length();
		ASpec[7] = A_bottom.length();
		//direct transport
		ASpec[8] = A_d;
		Float sumPdf = 0;
		for (int i = 0; i < 9; i++)
			sumPdf += ASpec[i];
		if (sumPdf == 0.0)
			return 0.0;
		for (int i = 0; i < 9; i++)
			ASpec[i] /= sumPdf;
		Float retVal = 0.0;
		for (int i = 0; i < 9; i++)
			retVal += pdfMLobe(i, theta_i, theta_o, phi) *
			pdfNLobe(i, h, theta_i, phi) *
			ASpec[i];
		retVal /= cos(theta_o);
		return retVal;
	}

	inline long double integrateLinear(long double a, long double b, long double c, long double d, long double e, long double f, long double h1, long double h2) const {
		// Common computations.
		long double exp1 = std::exp(-a * h1 + b);
		long double exp2 = std::exp(-a * h2 + b);
		long double a2 = a * a;
		long double a3 = a * a * a;
		long double a4 = a2 * a2;
		// Integrate cx^3*exp(-ax+b).
		long double p31 = -c * (a3 * h1 * h1 * h1 + 3.0 * a2 * h1 * h1 + 6.0 * a * h1 + 6) * exp1 / a4;
		long double p32 = -c * (a3 * h2 * h2 * h2 + 3.0 * a2 * h2 * h2 + 6.0 * a * h2 + 6) * exp2 / a4;
		// Integrate dx^2*exp(-ax+b).
		long double p21 = -d * (a2 * h1 * h1 + 2.0 * a * h1 + 2) * exp1 / a3;
		long double p22 = -d * (a2 * h2 * h2 + 2.0 * a * h2 + 2) * exp2 / a3;
		// Integrate ex*exp(-ax+b).
		long double p11 = -e * (a * h1 + 1) * exp1 / a2;
		long double p12 = -e * (a * h2 + 1) * exp2 / a2;
		// Integrate f*exp(-ax+b).
		long double p01 = -f * exp1 / a;
		long double p02 = -f * exp2 / a;
		return (p02 + p12 + p22 + p32) - (p01 + p11 + p21 + p31);
	}

	Vector3 integrateS(int p, long double kappa, long double f1, long double f2, long double s11, long double s12, long double s21, long double s22, long double avgDist1, long double avgDist2, Vector3 sigma_a_prime, long double sigma_e_m_prime, long double sigma_a_m_prime, long double D1, long double D2, long double h1, long double h2) const {
		long double x1 = std::pow(f1, (long double)(p - 1.0));
		long double y1 = 1.0 - f1;
		long double x2 = std::pow(f2, (long double)(p - 1.0));
		long double y2 = 1.0 - f2;
		long double k_x = (x2 - x1) / (h2 - h1);
		long double b_x = x1 - h1 * k_x;
		long double k_y = (y2 - y1) / (h2 - h1);
		long double b_y = y1 - h1 * k_y;
		long double k_D = (D2 - D1) / (h2 - h1);
		long double b_D = D1 - h1 * k_D;
		long double c = k_x * k_y * k_D;
		long double d = k_x * k_y * b_D + k_x * b_y * k_D + b_x * k_y * k_D;
		long double e = k_x * b_y * b_D + b_x * b_y * k_D + b_x * k_y * b_D;
		long double f = b_x * b_y * b_D;
		long double k_s1 = (s12 - s11) / (h2 - h1);
		long double b_s1 = s11 - h1 * k_s1;
		long double k_s2 = (s22 - s21) / (h2 - h1);
		long double b_s2 = s21 - h1 * k_s2;
		long double k_avgDist = (avgDist2 - avgDist1) / (h2 - h1);
		long double b_avgDist = avgDist1 - h1 * k_avgDist;
		Vector3 retVal;
		for (int i = 0; i < 3; i++) {
			// Note there's already a negative symbol before a.
			long double a = (2.0 * p - 1.0) * k_s1 * sigma_a_prime[i] +
				2.0 * (p - 1.0) * k_s2 * sigma_e_m_prime +
				k_avgDist * sigma_a_m_prime;
			long double b = -(((2.0 * p - 1.0) * b_s1 + (1.0 - kappa)) * sigma_a_prime[i] +
				2.0 * (p - 1.0) * b_s2 * sigma_e_m_prime +
				b_avgDist * sigma_a_m_prime);
			retVal[i] = integrateLinear(a, b, c, d, e, f, h1, h2);
		}
		return Float(0.5) * retVal;
	}

	Vector3 Ns(int p, Float phi, Float theta_i) const {
		if (kappa == 0.0)
			return Vector3(0.0, 0.0, 0.0);
		// Now using our analytic method.
		// Assume h is in [0, 1].
		const int SEGMENTS = 4;
		Float eta_prime = std::sqrt(eta * eta - sin(theta_i) * sin(theta_i)) / cos(theta_i);
		Float hMax = min(Float(1.0), kappa * eta_prime) - EPS;
		Vector3 retVal(0.0, 0.0, 0.0);
		for (int i = 0; i < SEGMENTS; i++) {
			Float h1 = (i + 0.0) / SEGMENTS;
			Float h2 = (i + 1.0) / SEGMENTS;
			h1 = std::sqrt(h1);
			h2 = std::sqrt(h2);
			h1 *= hMax;
			h2 *= hMax;
			Float gamma_i1 = asinSafe(h1);
			Float gamma_t1 = asinSafe(h1 / eta_prime);
			Float f1 = fresnelCuticle(gamma_i1, 1.0, eta_prime, layers);
			Float gamma_i2 = asinSafe(h2);
			Float gamma_t2 = asinSafe(h2 / eta_prime);
			Float f2 = fresnelCuticle(gamma_i2, 1.0, eta_prime, layers);
			Float theta_t = asinSafe(1.0 / eta * sin(theta_i));
			Vector3 sigma_a_prime = sigma_a / cos(theta_t);
			Float sigma_a_m_prime = sigma_a_m / cos(theta_t);
			Float sigma_s_m_prime = sigma_s_m / cos(theta_t);
			Float sigma_e_m_prime = sigma_a_m_prime + sigma_s_m_prime;
			// Travelling distance in the cortex and medulla, respectively.
			Float d11 = cos(gamma_t1), d21 = 0.0;
			Float h_prime1 = h1 / eta_prime;
			if (std::fabs(h_prime1) < std::fabs(kappa))
				d21 = std::sqrt(kappa * kappa - h_prime1 * h_prime1);
			d11 -= d21;
			Float d12 = cos(gamma_t2), d22 = 0.0;
			Float h_prime2 = h2 / eta_prime;
			if (std::fabs(h_prime2) < std::fabs(kappa))
				d22 = std::sqrt(kappa * kappa - h_prime2 * h_prime2);
			d12 -= d22;
			Float phi_query1, phi_query2, avgDist1, avgDist2, D1, D2;
			// Use phi.
			phi_query1 = phi - (gamma_t1 - gamma_i1) - (M_PI + 2.0 * gamma_t1) * (p - 1);
			regularizePhi(phi_query1);
			avgDist1 = kappa * getScatteredDist(sigma_s_m_prime * kappa, h_prime1 / kappa, g, phi_query1);
			D1 = getScatteredN(sigma_s_m_prime * kappa, h_prime1 / kappa, g, phi_query1);
			phi_query2 = phi - (gamma_t2 - gamma_i2) - (M_PI + 2.0 * gamma_t2) * (p - 1);
			regularizePhi(phi_query2);
			avgDist2 = kappa * getScatteredDist(sigma_s_m_prime * kappa, h_prime2 / kappa, g, phi_query2);
			D2 = getScatteredN(sigma_s_m_prime * kappa, h_prime2 / kappa, g, phi_query2);
			retVal += integrateS(p, kappa, f1, f2, d11, d12, d21, d22, avgDist1, avgDist2, sigma_a_prime, sigma_e_m_prime, sigma_a_m_prime, D1, D2, h1, h2);
			// Use -phi.
			phi_query1 = -phi - (gamma_t1 - gamma_i1) - (M_PI + 2.0 * gamma_t1) * (p - 1);
			regularizePhi(phi_query1);
			avgDist1 = kappa * getScatteredDist(sigma_s_m_prime * kappa, h_prime1 / kappa, g, phi_query1);
			D1 = getScatteredN(sigma_s_m_prime * kappa, h_prime1 / kappa, g, phi_query1);
			phi_query2 = -phi - (gamma_t2 - gamma_i2) - (M_PI + 2.0 * gamma_t2) * (p - 1);
			regularizePhi(phi_query2);
			avgDist2 = kappa * getScatteredDist(sigma_s_m_prime * kappa, h_prime2 / kappa, g, phi_query2);
			D2 = getScatteredN(sigma_s_m_prime * kappa, h_prime2 / kappa, g, phi_query2);
			retVal += integrateS(p, kappa, f1, f2, d11, d12, d21, d22, avgDist1, avgDist2, sigma_a_prime, sigma_e_m_prime, sigma_a_m_prime, D1, D2, h1, h2);
		}
		return retVal;
	}

	Vector3 NsMid(int p, Float hMinInput, Float hMaxInput, Float phi, Float theta_i) const {
		if (kappa == 0.0)
			return Vector3(0.0, 0.0, 0.0);
		// Now using our analytic method.
		// Assume h is in [0, 1].
		const int SEGMENTS = 4;
		Float eta_prime = std::sqrt(eta * eta - sin(theta_i) * sin(theta_i)) / cos(theta_i);
		Float hMax = min(Float(1.0), kappa * eta_prime) - EPS;
		Vector3 retVal(0.0, 0.0, 0.0);
		for (int i = 0; i < SEGMENTS; i++) {
			Float h1 = (i + 0.0) / SEGMENTS;
			Float h2 = (i + 1.0) / SEGMENTS;
			h1 = std::sqrt(h1);
			h2 = std::sqrt(h2);
			h1 *= hMax;
			h2 *= hMax;
			if (h2 <= hMinInput || h1 >= hMaxInput)
				continue;
			h1 = max(h1, hMinInput);
			h2 = min(h2, hMaxInput);
			Float gamma_i1 = asinSafe(h1);
			Float gamma_t1 = asinSafe(h1 / eta_prime);
			Float f1 = fresnelCuticle(gamma_i1, 1.0, eta_prime, layers);
			Float gamma_i2 = asinSafe(h2);
			Float gamma_t2 = asinSafe(h2 / eta_prime);
			Float f2 = fresnelCuticle(gamma_i2, 1.0, eta_prime, layers);
			Float theta_t = asinSafe(1.0 / eta * sin(theta_i));
			Vector3 sigma_a_prime = sigma_a / cos(theta_t);
			Float sigma_a_m_prime = sigma_a_m / cos(theta_t);
			Float sigma_s_m_prime = sigma_s_m / cos(theta_t);
			Float sigma_e_m_prime = sigma_a_m_prime + sigma_s_m_prime;
			// Travelling distance in the cortex and medulla, respectively.
			Float d11 = cos(gamma_t1), d21 = 0.0;
			Float h_prime1 = h1 / eta_prime;
			if (std::fabs(h_prime1) < std::fabs(kappa))
				d21 = std::sqrt(kappa * kappa - h_prime1 * h_prime1);
			d11 -= d21;
			Float d12 = cos(gamma_t2), d22 = 0.0;
			Float h_prime2 = h2 / eta_prime;
			if (std::fabs(h_prime2) < std::fabs(kappa))
				d22 = std::sqrt(kappa * kappa - h_prime2 * h_prime2);
			d12 -= d22;
			Float phi_query1, phi_query2, avgDist1, avgDist2, D1, D2;
			// Use phi.
			phi_query1 = phi - (gamma_t1 - gamma_i1) - (M_PI + 2.0 * gamma_t1) * (p - 1);
			regularizePhi(phi_query1);
			avgDist1 = kappa * getScatteredDist(sigma_s_m_prime * kappa, h_prime1 / kappa, g, phi_query1);
			D1 = getScatteredN(sigma_s_m_prime * kappa, h_prime1 / kappa, g, phi_query1);
			phi_query2 = phi - (gamma_t2 - gamma_i2) - (M_PI + 2.0 * gamma_t2) * (p - 1);
			regularizePhi(phi_query2);
			avgDist2 = kappa * getScatteredDist(sigma_s_m_prime * kappa, h_prime2 / kappa, g, phi_query2);
			D2 = getScatteredN(sigma_s_m_prime * kappa, h_prime2 / kappa, g, phi_query2);
			retVal += integrateS(p, kappa, f1, f2, d11, d12, d21, d22, avgDist1, avgDist2, sigma_a_prime, sigma_e_m_prime, sigma_a_m_prime, D1, D2, h1, h2);
			}
		return retVal / (hMaxInput - hMinInput) * 2.0;
	}

	inline long double erfMine(long double x) const {
		long double tau;
		long double t = 1.0 / (1.0 + 0.5 * std::fabs(x));
		long double k = -x * x -
			1.26551223 +
			1.00002368 * t +
			0.37409196 * std::pow(t, (long double) 2.0) +
			0.09678418 * std::pow(t, (long double) 3.0) -
			0.18628806 * std::pow(t, (long double) 4.0) +
			0.27886807 * std::pow(t, (long double) 5.0) -
			1.13520398 * std::pow(t, (long double) 6.0) +
			1.48851587 * std::pow(t, (long double) 7.0) -
			0.82215223 * std::pow(t, (long double) 8.0) +
			0.17087277 * std::pow(t, (long double) 9.0);
		tau = t * std::exp(k);
		if (x >= 0)
			return 1.0 - tau;
		else
			return tau - 1.0;
	}

	inline long double integrateR(long double f1, long double f2, long double d1, long double d2, long double h1, long double h2) const {
		const long double beta = beta_n;
		long double k_d = (d2 - d1) / (h2 - h1);
		long double b_d = d1 - h1 * k_d;
		long double k_A = (f2 - f1) / (h2 - h1);
		long double b_A = f1 - h1 * k_A;
		long double c = 1.0 / 2.0 / (std::sqrt(2.0 * M_PI) * beta);
		long double t = beta / (2.0 * k_d * k_d);
		long double erf_h2 = erfMine(d2 / (std::sqrt(2.0) * beta));
		long double exp_h2 = std::exp(-d2 * d2 / (2.0 * beta * beta));
		long double erf_h1 = erfMine(d1 / (std::sqrt(2.0) * beta));
		long double exp_h1 = std::exp(-d1 * d1 / (2.0 * beta * beta));
		long double I = t * (std::sqrt(2.0 * M_PI) * (erf_h2 - erf_h1) * (b_A * k_d - b_d * k_A) - 2.0 * k_A * beta * (exp_h2 - exp_h1));
		long double retVal = c * I;
		if (retVal < 0.0)
			return 0.0;
		if (std::isinf(retVal) || std::isinf(retVal) || std::isnan(retVal) || std::isnan(retVal))
			return 0.0;
		return retVal;
	}

	Float integrateR(Float f1, Float f2, Float Phi1, Float Phi2, Float phi, Float h1, Float h2) const {
		Float d1 = phi - Phi1;
		Float d2 = phi - Phi2;
		Float dd = d2 - d1;
		regularizePhi(d1);
		d2 = d1 + dd;
		if (d2 > M_PI) {
			return integrateR(f1, f2, d1, d2, h1, h2) +
				integrateR(f1, f2, d1 - M_PI * 2.0, d2 - M_PI * 2.0, h1, h2);
		}
		else if (d2 < -M_PI) {
			return integrateR(f1, f2, d1, d2, h1, h2) +
				integrateR(f1, f2, d1 + M_PI * 2.0, d2 + M_PI * 2.0, h1, h2);
		}
		return integrateR(f1, f2, d1, d2, h1, h2);
	}

	inline long double integrateQuad(long double a, long double b, long double c, long double d, long double e, long double f, long double h1, long double h2) const {
		if (a <= 0.0)
			return 0.0;
		if (std::fabs((2.0 * a * h1 - b) / (2.0 * std::sqrt(a))) > 5.0 &&
			std::fabs((2.0 * a * h2 - b) / (2.0 * std::sqrt(a))) > 5.0)
			return 0.0;
		long double x1 = erfMine((2.0 * a * h1 - b) / (2.0 * std::sqrt(a)));
		long double x2 = erfMine((2.0 * a * h2 - b) / (2.0 * std::sqrt(a)));
		long double y1 = std::exp(-a * h1 * h1 + b * h1 + c) * (2.0 * a * (d * h1 + e) + b * d);
		long double y2 = std::exp(-a * h2 * h2 + b * h2 + c) * (2.0 * a * (d * h2 + e) + b * d);
		long double p = std::sqrt(M_PI) / (8.0 * std::pow(a, (long double) 2.5)) * std::exp(c + (b * b) / (4.0 * a)) * (4.0 * a * a * f + 2.0 * a * (b * e + d) + b * b * d);
		long double q = 1.0 / (4.0 * a * a);
		if (std::isinf(p) || std::isinf(q) || std::isnan(p) || std::isnan(q))
			return 0.0;
		long double retVal = (p * x2 - q * y2) - (p * x1 - q * y1);
		return retVal < 0.0 ? 0.0 : retVal;
	}

	inline long double integrateQuad(long double a, long double b, long double c, long double e, long double f, long double h1, long double h2) const {
		if (a <= 0.0)
			return 0.0;
		if (std::fabs((2.0 * a * h1 - b) / (2.0 * std::sqrt(a))) > 5.0 &&
			std::fabs((2.0 * a * h2 - b) / (2.0 * std::sqrt(a))) > 5.0)
			return 0.0;
		long double x1 = erfMine((2.0 * a * h1 - b) / (2.0 * std::sqrt(a)));
		long double x2 = erfMine((2.0 * a * h2 - b) / (2.0 * std::sqrt(a)));
		long double y1 = std::exp(-a * h1 * h1 + b * h1 + c) * (2.0 * a * e);
		long double y2 = std::exp(-a * h2 * h2 + b * h2 + c) * (2.0 * a * e);
		long double p = std::sqrt(M_PI) / (8.0 * std::pow(a, (long double) 2.5)) * std::exp(c + (b * b) / (4.0 * a)) * (4.0 * a * a * f + 2.0 * a * b * e);
		long double q = 1.0 / (4.0 * a * a);
		if (std::isinf(p) || std::isinf(q) || std::isnan(p) || std::isnan(q))
			return 0.0;
		long double retVal = (p * x2 - q * y2) - (p * x1 - q * y1);
		return retVal < 0.0 ? 0.0 : retVal;
	}

	inline long double integrateQuad(long double a, long double b, long double c, long double f, long double h1, long double h2) const {
		if (a <= 0.0)
			return 0.0;
		if (std::fabs((2.0 * a * h1 - b) / (2.0 * std::sqrt(a))) > 5.0 &&
			std::fabs((2.0 * a * h2 - b) / (2.0 * std::sqrt(a))) > 5.0)
			return 0.0;
		long double x1 = erfMine((2.0 * a * h1 - b) / (2.0 * std::sqrt(a)));
		long double x2 = erfMine((2.0 * a * h2 - b) / (2.0 * std::sqrt(a)));
		long double y1 = 0.0;
		long double y2 = 0.0;
		long double p = std::sqrt(M_PI) / (8.0 * std::sqrt(a)) * std::exp(c + (b * b) / (4.0 * a)) * (4.0 * f);
		long double q = 1.0 / (4.0 * a * a);
		if (std::isinf(p) || std::isinf(q) || std::isnan(p) || std::isnan(q))
			return 0.0;
		long double retVal = (p * x2 - q * y2) - (p * x1 - q * y1);
		return retVal < 0.0 ? 0.0 : retVal;
	}

	inline Float integrateU(Float beta_p, int p, Float d, Float e, Float f, Float k_s1, Float b_s1, Float k_s2, Float b_s2, Float d1, Float d2, Float sigma_aa, Float sigma_ee, Float h1, Float h2) const {
		// Judge if \phi is between [\Phi_{min} - 3 \beta, \Phi_{max} + 3 \beta].
		Float dMin = std::min(d1, d2) - 3.0 * beta_p;
		Float dMax = std::max(d1, d2) + 3.0 * beta_p;
		if (!(dMin <= 0.0 && dMax >= 0.0))
			return 0.0;
		// Now calculate a, b, c;
		Float cc = 1.0 / 2.0 / (std::sqrt(2.0 * M_PI) * beta_p);
		Float k_d = (d2 - d1) / (h2 - h1);
		Float b_d = d1 - h1 * k_d;
		//        k_d = std::min(1000.0, std::max(1e-3, k_d));
		Float a = k_d * k_d / (2.0 * beta_p * beta_p);
		Float b = -(k_s1 * 2.0 * p * sigma_aa + k_s2 * 2.0 * p * sigma_ee + k_d * b_d / (beta_p * beta_p));
		Float c = -(b_s1 * 2.0 * p * sigma_aa + b_s2 * 2.0 * p * sigma_ee + b_d * b_d / (2.0 * beta_p * beta_p));
		if (a < 0.0) {
			return 0.0;
		}
		if (d == 0.0 && e == 0.0)
			return cc * integrateQuad(a, b, c, f, h1, h2);
		if (d == 0.0)
			return cc * integrateQuad(a, b, c, e, f, h1, h2);	
		return cc * integrateQuad(a, b, c, d, e, f, h1, h2);
	}

	Vector3 integrateU(int p, Float f1, Float f2, Float s11, Float s12, Float s21, Float s22, Vector3 sigma_a_prime, Float sigma_e_m_prime, Float Phi1, Float Phi2, Float phi, Float h1, Float h2) const {
		Float x1 = std::pow(f1, p - 1.0);
		Float y1 = std::pow(1.0 - f1, 2.0);
		Float x2 = std::pow(f2, p - 1.0);
		Float y2 = std::pow(1.0 - f2, 2.0);
		Float k_x = (x2 - x1) / (h2 - h1);
		Float b_x = x1 - h1 * k_x;
		Float k_y = (y2 - y1) / (h2 - h1);
		Float b_y = y1 - h1 * k_y;
		Float d = k_x * k_y;
		Float e = k_x * b_y + k_y * b_x;
		Float f = b_x * b_y;
		Float k_s1 = (s12 - s11) / (h2 - h1);
		Float b_s1 = s11 - h1 * k_s1;
		Float k_s2 = (s22 - s21) / (h2 - h1);
		Float b_s2 = s21 - h1 * k_s2;
		const Float &beta = beta_n;
		Float beta_p_sqr = beta * beta * (p + 1);
		Float beta_p = std::sqrt(beta_p_sqr);
		Float d1 = phi - Phi1;
		Float d2 = phi - Phi2;
		Float dd = d2 - d1;
		regularizePhi(d1);
		d2 = d1 + dd;
		if (d2 > M_PI) {
			return Vector3(integrateU(beta_p, p, d, e, f, k_s1, b_s1, k_s2, b_s2, d1, d2, sigma_a_prime[0], sigma_e_m_prime, h1, h2) +
				integrateU(beta_p, p, d, e, f, k_s1, b_s1, k_s2, b_s2, d1 - M_PI * 2.0, d2 - M_PI * 2.0, sigma_a_prime[0], sigma_e_m_prime, h1, h2),
				integrateU(beta_p, p, d, e, f, k_s1, b_s1, k_s2, b_s2, d1, d2, sigma_a_prime[1], sigma_e_m_prime, h1, h2) +
				integrateU(beta_p, p, d, e, f, k_s1, b_s1, k_s2, b_s2, d1 - M_PI * 2.0, d2 - M_PI * 2.0, sigma_a_prime[1], sigma_e_m_prime, h1, h2),
				integrateU(beta_p, p, d, e, f, k_s1, b_s1, k_s2, b_s2, d1, d2, sigma_a_prime[2], sigma_e_m_prime, h1, h2) +
				integrateU(beta_p, p, d, e, f, k_s1, b_s1, k_s2, b_s2, d1 - M_PI * 2.0, d2 - M_PI * 2.0, sigma_a_prime[2], sigma_e_m_prime, h1, h2));
		}
		else if (d2 < -M_PI) {
			return Vector3(integrateU(beta_p, p, d, e, f, k_s1, b_s1, k_s2, b_s2, d1, d2, sigma_a_prime[0], sigma_e_m_prime, h1, h2) +
				integrateU(beta_p, p, d, e, f, k_s1, b_s1, k_s2, b_s2, d1 + M_PI * 2.0, d2 + M_PI * 2.0, sigma_a_prime[0], sigma_e_m_prime, h1, h2),
				integrateU(beta_p, p, d, e, f, k_s1, b_s1, k_s2, b_s2, d1, d2, sigma_a_prime[1], sigma_e_m_prime, h1, h2) +
				integrateU(beta_p, p, d, e, f, k_s1, b_s1, k_s2, b_s2, d1 + M_PI * 2.0, d2 + M_PI * 2.0, sigma_a_prime[1], sigma_e_m_prime, h1, h2),
				integrateU(beta_p, p, d, e, f, k_s1, b_s1, k_s2, b_s2, d1, d2, sigma_a_prime[2], sigma_e_m_prime, h1, h2) +
				integrateU(beta_p, p, d, e, f, k_s1, b_s1, k_s2, b_s2, d1 + M_PI * 2.0, d2 + M_PI * 2.0, sigma_a_prime[2], sigma_e_m_prime, h1, h2));
		}
		return Vector3(integrateU(beta_p, p, d, e, f, k_s1, b_s1, k_s2, b_s2, d1, d2, sigma_a_prime[0], sigma_e_m_prime, h1, h2),
			integrateU(beta_p, p, d, e, f, k_s1, b_s1, k_s2, b_s2, d1, d2, sigma_a_prime[1], sigma_e_m_prime, h1, h2),
			integrateU(beta_p, p, d, e, f, k_s1, b_s1, k_s2, b_s2, d1, d2, sigma_a_prime[2], sigma_e_m_prime, h1, h2));
	}

	Vector3 Nu(int p, Float phi, Float theta_i) const {
		// Now using our analytic method.
		// Assume h is in [0, 1].
		const int SEGMENTS = 5;
		Float eta_prime = std::sqrt(eta * eta - sin(theta_i) * sin(theta_i)) / cos(theta_i);
		Vector3 retVal(0.0, 0.0, 0.0);
		for (int i = 0; i < SEGMENTS; i++) {
			Float h1 = (i + 0.0) / SEGMENTS;
			Float h2 = (i + 1.0) / SEGMENTS;
			h1 *= 1.0 - EPS;
			h2 *= 1.0 - EPS;
			h1 = std::sqrt(h1);
			h2 = std::sqrt(h2);
			Float gamma_i1 = asinSafe(h1);
			Float gamma_t1 = asinSafe(h1 / eta_prime);
			Float f1 = fresnelCuticle(gamma_i1, 1.0, eta_prime, layers);
			Float Phi1 = 2.0 * p * gamma_t1 - 2.0 * gamma_i1 + p * M_PI;
			Float gamma_i2 = asinSafe(h2);
			Float gamma_t2 = asinSafe(h2 / eta_prime);
			Float f2 = fresnelCuticle(gamma_i2, 1.0, eta_prime, layers);
			Float Phi2 = 2.0 * p * gamma_t2 - 2.0 * gamma_i2 + p * M_PI;
			if (p == 0) {
				retVal += Vector3(1.0, 1.0, 1.0) * integrateR(f1, f2, Phi1, Phi2, phi, h1, h2);
				retVal += Vector3(1.0, 1.0, 1.0) * integrateR(f1, f2, Phi1, Phi2, -phi, h1, h2);
			}
			else {
				Float theta_t = asinSafe(1.0 / eta * sin(theta_i));
				Vector3 sigma_a_prime = sigma_a / cos(theta_t);
				Float sigma_a_m_prime = sigma_a_m / cos(theta_t);
				Float sigma_s_m_prime = sigma_s_m / cos(theta_t);
				Float sigma_e_m_prime = sigma_a_m_prime + sigma_s_m_prime;

				// Travelling distance in the cortex and medulla, respectively.
				Float d11 = cos(gamma_t1), d21 = 0.0;
				Float h_prime1 = h1 / eta_prime;
				if (std::fabs(h_prime1) < std::fabs(kappa))
					d21 = std::sqrt(kappa * kappa - h_prime1 * h_prime1);
				d11 -= d21;
				Float d12 = cos(gamma_t2), d22 = 0.0;
				Float h_prime2 = h2 / eta_prime;
				if (std::fabs(h_prime2) < std::fabs(kappa))
					d22 = std::sqrt(kappa * kappa - h_prime2 * h_prime2);
				d12 -= d22;
				retVal += integrateU(p, f1, f2, d11, d12, d21, d22, sigma_a_prime, sigma_e_m_prime, Phi1, Phi2, phi, h1, h2);
				retVal += integrateU(p, f1, f2, d11, d12, d21, d22, sigma_a_prime, sigma_e_m_prime, Phi1, Phi2, -phi, h1, h2);
			}
		}
		if (std::isnan(retVal[0]) || std::isnan(retVal[1]) || std::isnan(retVal[2]))
			retVal = Vector3(0.0f, 0.0f, 0.0f);
		if (std::isinf(retVal[0]) || std::isinf(retVal[1]) || std::isinf(retVal[2]))
			retVal = Vector3(0.0f, 0.0f, 0.0f);
		for (int i = 0; i < 3; i++)
			retVal[i] = std::min(std::max(Float(0.0f), retVal[i]), Float(1e3f));
		return retVal;
	}

	Vector3 NuMid(int p, Float hMinInput, Float hMaxInput, Float phi, Float theta_i) const {
		// Now using our analytic method.
		// Assume h is in [0, 1].
		//        Assert(hMinInput >= 0.0 && hMaxInput >= 0.0);
		const int SEGMENTS = 5;
		Float eta_prime = std::sqrt(eta * eta - sin(theta_i) * sin(theta_i)) / cos(theta_i);
		Vector3 retVal(0.0, 0.0, 0.0);
		for (int i = 0; i < SEGMENTS; i++) {
			Float h1 = (i + 0.0) / SEGMENTS;
			Float h2 = (i + 1.0) / SEGMENTS;
			h1 *= 1.0 - EPS;
			h2 *= 1.0 - EPS;
			h1 = std::sqrt(h1);
			h2 = std::sqrt(h2);
			if (h2 <= hMinInput || h1 >= hMaxInput)
				continue;
			h1 = max(h1, hMinInput);
			h2 = min(h2, hMaxInput);
			Float gamma_i1 = asinSafe(h1);
			Float gamma_t1 = asinSafe(h1 / eta_prime);
			Float f1 = fresnelCuticle(gamma_i1, 1.0, eta_prime, layers);
			Float Phi1 = 2.0 * p * gamma_t1 - 2.0 * gamma_i1 + p * M_PI;
			Float gamma_i2 = asinSafe(h2);
			Float gamma_t2 = asinSafe(h2 / eta_prime);
			Float f2 = fresnelCuticle(gamma_i2, 1.0, eta_prime, layers);
			Float Phi2 = 2.0 * p * gamma_t2 - 2.0 * gamma_i2 + p * M_PI;
			if (p == 0) {
				retVal += Vector3(1.0, 1.0, 1.0) * integrateR(f1, f2, Phi1, Phi2, phi, h1, h2);
			}
			else {
				Float theta_t = asinSafe(1.0 / eta * sin(theta_i));
				Vector3 sigma_a_prime = sigma_a / cos(theta_t);
				Float sigma_a_m_prime = sigma_a_m / cos(theta_t);
				Float sigma_s_m_prime = sigma_s_m / cos(theta_t);
				Float sigma_e_m_prime = sigma_a_m_prime + sigma_s_m_prime;
				// Travelling distance in the cortex and medulla, respectively.
				Float d11 = cos(gamma_t1), d21 = 0.0;
				Float h_prime1 = h1 / eta_prime;
				if (std::fabs(h_prime1) < std::fabs(kappa))
					d21 = std::sqrt(kappa * kappa - h_prime1 * h_prime1);
				d11 -= d21;
				Float d12 = cos(gamma_t2), d22 = 0.0;
				Float h_prime2 = h2 / eta_prime;
				if (std::fabs(h_prime2) < std::fabs(kappa))
					d22 = std::sqrt(kappa * kappa - h_prime2 * h_prime2);
				d12 -= d22;
				retVal += integrateU(p, f1, f2, d11, d12, d21, d22, sigma_a_prime, sigma_e_m_prime, Phi1, Phi2, phi, h1, h2);
				//                retVal += integrateU(p, f1, f2, d11, d12, d21, d22, sigma_a_prime, sigma_e_m_prime, Phi1, Phi2, -phi, h1, h2);
			}
		}
		if (std::isnan(retVal[0]) || std::isnan(retVal[1]) || std::isnan(retVal[2]))
			retVal = Vector3(0.0f, 0.0f, 0.0f);
		if (std::isinf(retVal[0]) || std::isinf(retVal[1]) || std::isinf(retVal[2]))
			retVal = Vector3(0.0f, 0.0f, 0.0f);
		for (int i = 0; i < 3; i++)
			retVal[i] = std::min(std::max(Float(0.0f), retVal[i]), Float(1e3f));
		//        for (int i = 0; i < 3; i++)
		//            printf("%f %f %f\n", retVal[0], retVal[1], retVal[2]);
		return retVal / (hMaxInput - hMinInput) * 2.0;
	}

	Vector3 Ne(int p, Float theta_i, Float phi) const {
		Vector3 retVal(0.0, 0.0, 0.0);
		//get the value from spline
		switch (p) {
		case 0: // top
			for (int i = 0; i < 3; i++)
			{
				retVal[i] = EN[i].getval(phi)*A_top[i];
			}
			
		case 1: // middle
			for (int i = 0; i < 3; i++)
			{
				retVal[i] = EN[i].getval(phi)*A_middle[i];
			}
		case 2: // bottom
			for (int i = 0; i < 3; i++)
			{
				retVal[i] = EN[i].getval(phi)*A_bottom[i];
			}
		}
		return retVal;
	}

	Vector3 Nd(Float phi) const {
		Vector3 retVal(0.0, 0.0, 0.0);
		Float val = A_d*G(normalizephi(phi-M_PI), Rd);
		retVal = Vector3(val, val, val);
		return retVal;
	}

};

MTS_NAMESPACE_END

#endif
